<?
////////////////////////////////////////////////////////////////////////
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
////////////////////////////////////////////////////////////////////////
class CardDB {
	private $setDB;
	private $artDB;
	private $titleToCards = array();
	private $titleToFlavors = array();
	private $colorCombination = array("WUB" => "WUB","WUR" => "RWU","WUG" => "GWU","WBR" => "WBR","WBG" => "BGW","WRG" => "RGW","UBR" => "UBR","UBG" => "GUB","URG" => "URG","BRG" => "BRG",);

	public function __construct (SetDB $setDB, ArtDB $artDB) {
		global $config;

		$this->setDB = $setDB;
		$this->artDB = $artDB;

		// Load english cards.
		echo 'Loading card data';
		$file = file_get_contents('data/cards.json', 'r');
		if (!$file) error('Unable to open file: data/cards.json');
		$jsonFile = json_decode($file, true);
		foreach($jsonFile as $setName => $set){
			foreach($set["cards"] as $key => $cardJson){
				$card = $this->jsonToCard($cardJson, $setName, $set["cards"]);
				if($card != null){
					$card->set = $setDB->normalize($card->set);
					if (!$card->set) continue;

					$title = strtolower($card->title);
					if (!@$this->titleToCards[$title]) $this->titleToCards[$title] = array();
					$this->titleToCards[(string)$title][] = $card;
				}
			}
		}
		echo "\n";

		// Load foreign card data.
		$language = strtolower($config['output.language']);
		if ($language && $language != 'english') {
			echo "Loading $language card data";
			$file = fopen_utf8("data/cards-$language.csv", 'r');
			if (!$file) error("Unable to open file: data/cards-$language.csv");
			$i = 0;
			while (($row = fgetcsv($file, 6000, ',')) !== FALSE) {
				if ($i++ % 400 == 0) echo '.';
				// Overwrite some of the english card values with the foreign values.
				$englishTitle = strtolower($row[0]);

				$cards = @$this->titleToCards[(string)$englishTitle];
				if (!$cards){
					//print_r($row);
					echo "\nError matching card data for card: $row[0]";
					continue; // Skip errors
					}
				foreach ($cards as $card)
					CardDB::applyLanguageRowToCard($row, $card);
			}
			fclose($file);
			echo "\n";
			if (!$config['output.english.flavor.text']) {
				echo "Loading $language card flavor data";
				$file = fopen_utf8("data/cards-$language-flavor.csv", 'r');
				if (!$file)
					echo "\nNo localized flavor for language: $language";
				else {
					$i = 0;
					while (($row = fgetcsv($file, 6000, ',')) !== FALSE) {
						if ($i++ % 400 == 0) echo '.';
						// Overwrite some of the english card values with the foreign values.
						$englishTitle = strtolower($row[0]);
						$cards = @$this->titleToCards[(string)$englishTitle];
						if (!$cards){
							//print_r($row);
							echo "\nError matching card flavor for card: $row[0]";
							continue; // Skip errors.
							}
						// Find the card from needed edition and apply localized flavor.
						foreach ($cards as $card) {
							if ($card->set == $setDB->normalize($row[1])) {
								$card->flavor = $row[2];
								break;
							}
						}
					}
					fclose($file);
					echo "\n";
				}
			}
		}
	}

	public function getCard ($title, $set = null, $pic = null) {
		global $config;

		$cards = @$this->titleToCards[(string)strtolower($title)];
		if (!$cards) return null;
		if (!$set) {
			if ($config['card.set.random']) {
				// Pick a random set.
				$sets = array();
				foreach ($cards as $card)
					$sets[$card->set] = true;
				$set = array_rand($sets);
			} else {
				// Find earliest set.
				$lowest = 999999;
				foreach ($cards as $card) {
					$ordinal = $this->setDB->getOrdinal($card->set);
					if ($ordinal < $lowest) {
						$lowest = $ordinal;
						$set = $card->set;
					}
				}
			}
		} else
			$set = $this->setDB->normalize($set);

		$chosenCard = null;
		if ($pic) {
			// Find the specific picture number.
			foreach ($cards as $card) {
				if ($card->set != $set) continue;
				if ($card->pic == $pic) {
					$chosenCard = $card;
					break;
				}
			}
		} else {
			// Randomly pick a card in the set.
			$cardsInSet = array();
			foreach ($cards as $card) {
				if ($card->set != $set) continue;
				$cardsInSet[] = $card;
			}
			if (count($cardsInSet) > 0) $chosenCard = $cardsInSet[array_rand($cardsInSet)];
		}
		if (!$chosenCard) return null;

		// Card was found.
		return $this->configureCard($chosenCard);
	}

	public function getSets ($cardTitle) {
		$cards = @$this->titleToCards[(string)strtolower($cardTitle)];
		if (!$cards) return null;
		$sets = array();
		foreach ($cards as $card)
			$sets[$card->set] = true;
		return array_keys($sets);
	}

	public function getCards ($title) {
		$cards = @$this->titleToCards[(string)strtolower($title)];
		if (!$cards) return null;
		$configuredCards = array();
		foreach ($cards as $card)
			$configuredCards[] = $this->configureCard($card);
		return $configuredCards;
	}

	public function configureCard ($card) {
		global $config;

		$card = clone $card;

		if ($config["card.flavor.random"]) {
			// Pick a flavor text that hasn't been picked yet.
			$flavors = @$this->titleToFlavors[(string)$card->title];
			if (!$flavors || count($flavors) == 0) {
				$cardWithSameTitle = @$this->titleToCards[(string)strtolower($card->title)];
				foreach ($cardWithSameTitle as $cardWithSameTitle)
					if ($cardWithSameTitle->flavor) $flavors[] = $cardWithSameTitle->flavor;
			}
			if (count($flavors) > 0) {
				$index = array_rand($flavors);
				$card->flavor = $flavors[$index];
				array_splice($flavors, $index, 1);
			}
		}
		
		// Find art image.
		$card->artFileName = $this->artDB->getArtFileName($card->title, $card->set, $card->pic, $card->imageName);

		// Artist and copyright.
		if ($config['card.artist.and.copyright'])
			$card->copyright = $config['card.copyright'] . ' ' . $card->collectorNumber;
		else {
			$card->artist = null;
			$card->copyright = null;
		}

		if (!$config['card.reminder.text']) $card->legal = preg_replace('/#\\(.*?\\)#/', '', $card->legal);

		return $card;
	}

	public function getAllCardTitles () {
		$titles = array();
		foreach ($this->titleToCards as $title => $cards)
			$titles[$title] = true;
		return array_keys($titles);
	}

	static public function rowToCard ($row) {
		$card = new Card();
		$card->title = (string)$row[0];
		$card->set = (string)$row[1];
		$card->color = (string)$row[2];
		$card->type = (string)$row[3];
		$card->englishType = $card->type;
		$card->pt = (string)str_replace('\\', '/', $row[4]);
		$card->flavor = (string)$row[5];
		$card->rarity = (string)$row[6];
		$card->cost = (string)$row[7];
		$card->legal = (string)str_replace("\r\n", "\n", $row[8]);
		$card->pic = (string)@$row[9];
		$card->artist = (string)@$row[10];
		$card->collectorNumber = (string)str_replace('\\', '/', @$row[11]);
		return $card;
	}
	
	public function jsonToCard ($jsonCard, $set, $wholeSet) {
		if($jsonCard["layout"] == "token")
			return null;
		$card = new Card();
		$card->layout = $jsonCard["layout"];
		$card->title = $jsonCard["name"];
		
		//Get other flip card information
		if($card->layout == "flip"){
			if(isset($jsonCard["number"]) && substr($jsonCard["number"], -1) == "b") //Skip second side of card
				return null;
			foreach($jsonCard["names"] as $name){
				if($card->title != $name)
					$card->transformedTitle = $name;
			}
			foreach($wholeSet as $key => $jsonCardItem){
				if($jsonCardItem["name"] == $card->transformedTitle){
					if(isset($jsonCardItem["text"]))
						$card->transformedLegal = $jsonCardItem["text"];
					$card->transformedType = $jsonCardItem["type"];
					if(isset($jsonCardItem["power"]))
						$card->transformedPT = $jsonCardItem["power"]."/".$jsonCardItem["toughness"];
				}
			}
		}
		if($card->layout == "split"){
			foreach($jsonCard["names"] as $name){
				if($card->title != $name){
					$card->transformedTitle = $name;
					$card->otherSplitHalf = $this->getCard($card->transformedTitle, $set);
				}
			}
		}
			
			
		$card->set = $set;
		if(isset($jsonCard["colors"])){
			$colorWheel = "WUBRG";
			$dblColorWheel = "WUBRGWUBRG";
			if(in_array("White",$jsonCard["colors"]))
				$card->colors[] = "W";
			if(in_array("Blue",$jsonCard["colors"]))
				$card->colors[] = "U";
			if(in_array("Black",$jsonCard["colors"]))
				$card->colors[] = "B";
			if(in_array("Red",$jsonCard["colors"]))
				$card->colors[] = "R";
			if(in_array("Green",$jsonCard["colors"]))
				$card->colors[] = "G";
			
			//Sort colors according to magic rules
			if(count($card->colors) == 2)
				if(strpos($colorWheel, $card->colors[1]) - strpos($colorWheel, $card->colors[0]) >= 3)
					$card->colors = array($card->colors[1], $card->colors[0]);
			if(count($card->colors) == 3){
				$aggColors = $this->colorCombination[implode('',$card->colors)];
				$card->colors = array($aggColors[0], $aggColors[1], $aggColors[2]);
			}
			if(count($card->colors) == 4)
				for($i=0; strlen($colorWheel) < $i; $i++){
					if(!in_array($colorWheel[$i], $card->colors)){
						$card->colors = array($dblColorWheel[$i],$dblColorWheel[$i+1],$dblColorWheel[$i+2],$dblColorWheel[$i+3]);
						break;
					}
				}
				
		}
		$card->type = $jsonCard["type"];
		$card->englishType = $jsonCard["type"];
		
		if(isset($jsonCard["power"]))
			$card->pt = $jsonCard["power"]."/".$jsonCard["toughness"];
		if(in_array("Planeswalker",$jsonCard["types"]) && isset($jsonCard["loyalty"])){
			$card->loyalty = $jsonCard["loyalty"];
		}
		
		if(isset($jsonCard["watermark"]))
			$card->watermark = $jsonCard["watermark"];
			
		if(isset($jsonCard["flavor"]))
			$card->flavor = $jsonCard["flavor"];
		switch($jsonCard["rarity"]){
			case "Common" : $card->rarity = "C";break;
			case "Uncommon" : $card->rarity = "U";break;
			case "Rare" : $card->rarity = "R";break;
			case "Mythic Rare" : $card->rarity = "M";break;
			case "Basic Land" : $card->rarity = "C";break;
			default : if($jsonCard["rarity"] == "Special" && $set == "TSB") $card->rarity = "R";break;
		}
		if(isset($jsonCard["manaCost"])){
			$card->cost = str_replace("/", "", $jsonCard["manaCost"]);
		}
		if(isset($jsonCard["text"])){
			$jsonCard["text"] = str_replace(array('(', ')'), array('#(',')#'), $jsonCard["text"]);
			$card->legal = preg_replace(array('/\{B\/G\}/','/\{B\/R\}/','/\{G\/U\}/','/\{G\/W\}/','/\{R\/G\}/','/\{R\/W\}/','/\{U\/B\}/','/\{U\/R\}/','/\{W\/B\}/','/\{W\/U\}/','/\{2\/U\}/','/\{2\/B\}/','/\{2\/W\}/','/\{2\/R\}/','/\{2\/G\}/','/\{P\/W\}/','/\{P\/R\}/','/\{P\/G\}/','/\{P\/U\}/','/\{P\/B\}/'), array('{BG}','{BR}','{GU}','{GW}','{RG}','{RW}','{UB}','{UR}','{WB}','{WU}','{2U}','{2B}','{2W}','{2R}','{2G}','{PW}','{PR}','{PG}','{PU}','{PB}'), $jsonCard["text"]);
		}
		if(substr($card->legal,0,8) == "Level up")
			$card->layout = "leveler";
			
		if(isset($jsonCard["variations"])){
			if(is_numeric(substr($jsonCard["imageName"],-1)) && !is_numeric(substr($jsonCard["imageName"],-2,1)))
				$card->pic = substr($jsonCard["imageName"],-1); //forest1
			else if(is_numeric(substr($jsonCard["imageName"],-2)))
				$card->pic = substr($jsonCard["imageName"],-2); //forest12
			else if(!is_numeric(substr($jsonCard["imageName"],-1)) && is_numeric(substr($jsonCard["imageName"],-2,1)))
				$card->pic = substr($jsonCard["imageName"],-2); //forest1a
			
		}
		$card->imageName = $jsonCard["imageName"];
		$card->artist = $jsonCard["artist"];
		if(isset($jsonCard["number"]))
			$card->collectorNumber = $jsonCard["number"];
		if(!is_numeric(substr($card->collectorNumber, -1))) //For flip and split cards
			$card->collectorNumber = substr($card->collectorNumber, 0,-1);
		return $card;
	}

	static public function cardToRow ($card) {
		$row = array();
		$row[0] = (string)$card->title;
		$row[1] = (string)$card->set;
		$row[2] = (string)$card->color;
		$row[3] = (string)$card->type;
		$row[4] = (string)str_replace('/', '\\', $card->pt);
		$row[5] = (string)$card->flavor;
		$row[6] = (string)$card->rarity;
		$row[7] = (string)$card->cost;
		$row[8] = (string)$card->legal;
		$row[9] = (string)$card->pic;
		$row[10] = (string)$card->artist;
		$row[11] = (string)str_replace('/', '\\', $card->collectorNumber);
		return $row;
	}

	static public function applyLanguageRowToCard ($row, $card) {
		global $config;
		if($row[1]){
			$cardName = preg_replace("/ ?\(\d\)/", "", $row[1]);
			$card->setDisplayTitle($cardName);
		}
		else
			$card->setDisplayTitle($row[0]);
		$card->type = $row[2];
		if(@$row[3])
			$card->legal = str_replace("\r\n", "\n", $row[3]);
		else
			$card->legal = "";
		if (!$config['output.english.flavor.text']) $card->flavor = @$row[4];
		if ($config['output.english.title.on.translated.card']) $card->artist = $card->title;
	}

	static public function cardToLanguageRow ($card) {
		$row = array();
		$row[0] = $card->title;
		$row[1] = $card->getDisplayTitle();
		$row[2] = $card->type;
		$row[3] = $card->legal;
		$row[4] = $card->flavor;
		return $row;
	}

	public function resetUsedFlavors () {
		$this->titleToFlavors = array();
	}
}

?>
